#!/bin/sh
#ifdef __USAGE
#%C    - save TRAKT gate configuration.
#
#%C    [-n <node_num>] [<source>|<source> <destination>]
#Where:
#      node_num      number of the saving node. If node = 0,
#                    node number will get from NODE environment variable.
#      source        gate directory. Default is "/".
#      destination   directory to save settings. Default is "/dos/a".
#Notes:
#      Script is interactive and read stdin for user answers.
#Examples:
#      %C /mnt/hd0t77 /dos/a
#         - save configuration from first partition of the
#           first hard disk to /dos/a/
#endif


if [ "-$1" = "--n" ]; then
	GATE_NODE="$2"
	[ -z "$GATE_NODE" ] && { echo "Option '-n' needs parameter."; exit 1; }
	shift 2
fi

S_DIR=${1:-/}
D_DIR=${2:-/dos/a}

S_DIR="$(fullpath -t "$S_DIR")"

if [ ! -f "$S_DIR/boot/build/hard."[0-9]* -o ! -f "$S_DIR/etc/hosts" ]; then
    echo "\"$S_DIR\" is not a gate directory!"
    exit 1
fi

if [ -z "$GATE_NODE" ]; then
	GATE_NODE=$(ls -1 "$S_DIR/boot/build/hard."[0-9]*|awk 'BEGIN {FS="."} END {print $2}')
elif [ "$GATE_NODE" = "0" ]; then
	GATE_NODE="$NODE"
fi

GATE_HOST=$(grep "192.168.0.$GATE_NODE " "$S_DIR/etc/hosts"|awk '{ printf("%s", $2); }')
D_DIR="$D_DIR/${GATE_HOST}_${GATE_NODE}"

echo "Source:      $S_DIR"
echo "Destination: $D_DIR"
echo "Is it true? [yn]\c"
read resp

case $resp in
	y*|Y*)
	;;
	*)
		exit 1
	;;
esac

mkdir "$D_DIR" || { echo "Can't create directory \"$D_DIR\""; exit 1; }
#cfont -f 3 koi8-8x8
GATE_ID=$(grep "/usr/ucb/ifconfig en1 .* up" "$S_DIR/usr/ucb/runtcp"|awk '{ print $3; }')
GATE_ROUND=$(grep ".* - addr" "$S_DIR/etc/tract.log"|cut -f1 -d" - ")
echo "Enter gate name: \c"
read GATE_NAME
echo "Enter gate round (0xNNNN): \c"
read GATE_ROUND
echo "Enter gate place: \c"
read GATE_PLACE
GATE_PRIO="$(cat "$S_DIR/etc/hosts" | awk -v answ=$GATE_NODE '{ if (NR == answ) print $8}')"
GATE_IP="$(cat "$S_DIR/etc/hosts" | awk -v answ=$GATE_NODE '{ if (NR == answ) print $1}')"
echo "#\n# \"TRAKT\" gate metadata file.\n#\n" > "$D_DIR/metadata.txt"
echo "GATE_ID=\"$GATE_ID\"\nGATE_NAME=\"$GATE_NAME\"\nGATE_ROUND=\"$GATE_ROUND\"\nGATE_PRIO=\"$GATE_PRIO\"" >> "$D_DIR/metadata.txt"
echo "GATE_PLACE=\"$GATE_PLACE\"\nGATE_IP=\"$GATE_IP\"\nGATE_NODE=\"$GATE_NODE\"" >> "$D_DIR/metadata.txt"
cd "$S_DIR"
tar -cf - etc/config/sysinit.$GATE_NODE boot/build/hard.$GATE_NODE* \
          etc/tract.log \
          usr/ucb/cngcr usr/ucb/runtcp id \
          usr/check/[!v]* /usr/check/var/root.log | \
	gzip --best -c > "$D_DIR/data.tgz"

sync
