#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>

#ifdef __QNXNTO__
 #include <sys/dcmd_mixer.h>
 #define ioctl(a,b,c)	devctl(a,b,c,sizeof(*(c)),NULL)
 #define AUDIO_DEVICE	"/dev/audio"
#else
 #include <sys/audio.h>
 #define AUDIO_DEVICE	"/dev/dsp"
#endif


#ifdef __USAGE
%C	- Play audio files through /dev/dsp

%C	file...
#endif

static size_t filesize(FILE *);

u_long samples, samplerate, samplechannels, samplebits;
typedef int (*pfi)();
FILE *audio;
int	fd;

size_t	playsize;

xgetc(void)
{
    return fgetc(stdin);
}

xread(void *buffer, int length)
{
    if (fread(buffer, 1, length, stdin) != length)
		return 0;

    return length;
}

xrewind(void)
{
    fseek(stdin, 0, SEEK_SET);
}

xskip(long offset)
{
    fseek(stdin, offset, SEEK_CUR);
}

xtag(char *tag)
{
    struct riff {
		char tag[4];
		long length;
    } buffer;

    while (xread(&buffer, sizeof buffer)) {
		if (strncmp(tag, buffer.tag, sizeof buffer.tag) == 0)
		    return buffer.length;
		xskip(buffer.length);
    }

    return 0;
}

aul_read(void)
{
    if (samples == 0 || samples--)
		return xgetc();

    return EOF;
}

auu_read(void)
{
    if (samples == 0 || samples--) {
		static unsigned short ulaw_table[256] = {
	    -32124, -31100, -30076, -29052, -28028, -27004, -25980, -24956,
	    -23932, -22908, -21884, -20860, -19836, -18812, -17788, -16764,
	    -15996, -15484, -14972, -14460, -13948, -13436, -12924, -12412,
	    -11900, -11388, -10876, -10364,  -9852,  -9340,  -8828,  -8316,
	     -7932,  -7676,  -7420,  -7164,  -6908,  -6652,  -6396,  -6140,
	     -5884,  -5628,  -5372,  -5116,  -4860,  -4604,  -4348,  -4092,
	     -3900,  -3772,  -3644,  -3516,  -3388,  -3260,  -3132,  -3004,
	     -2876,  -2748,  -2620,  -2492,  -2364,  -2236,  -2108,  -1980,
	     -1884,  -1820,  -1756,  -1692,  -1628,  -1564,  -1500,  -1436,
	     -1372,  -1308,  -1244,  -1180,  -1116,  -1052,   -988,   -924,
	      -876,   -844,   -812,   -780,   -748,   -716,   -684,   -652,
	      -620,   -588,   -556,   -524,   -492,   -460,   -428,   -396,
	      -372,   -356,   -340,   -324,   -308,   -292,   -276,   -260,
	      -244,   -228,   -212,   -196,   -180,   -164,   -148,   -132,
	      -120,   -112,   -104,    -96,    -88,    -80,    -72,    -64,
	       -56,    -48,    -40,    -32,    -24,    -16,     -8,      0,
	     32124,  31100,  30076,  29052,  28028,  27004,  25980,  24956,
	     23932,  22908,  21884,  20860,  19836,  18812,  17788,  16764,
	     15996,  15484,  14972,  14460,  13948,  13436,  12924,  12412,
	     11900,  11388,  10876,  10364,   9852,   9340,   8828,   8316,
	      7932,   7676,   7420,   7164,   6908,   6652,   6396,   6140,
	      5884,   5628,   5372,   5116,   4860,   4604,   4348,   4092,
	      3900,   3772,   3644,   3516,   3388,   3260,   3132,   3004,
	      2876,   2748,   2620,   2492,   2364,   2236,   2108,   1980,
	      1884,   1820,   1756,   1692,   1628,   1564,   1500,   1436,
	      1372,   1308,   1244,   1180,   1116,   1052,    988,    924,
	       876,    844,    812,    780,    748,    716,    684,    652,
	       620,    588,    556,    524,    492,    460,    428,    396,
	       372,    356,    340,    324,    308,    292,    276,    260,
	       244,    228,    212,    196,    180,    164,    148,    132,
	       120,    112,    104,     96,     88,     80,     72,     64,
			56,     48,     40,     32,     24,     16,      8,      0
		};
		static int cont, ch;

		if (samplebits == 16) {
		    if ((++cont & 1) == 0)
				return ch >> 8;
		}

		if ((ch = xgetc()) == EOF) {
		    cont = 0;
		    return EOF;
		}

		ch = ulaw_table[(u_char) ch];

		if (samplebits == 8)
		    return (ch >> 8) ^ 128;

		return ch;
    }

    return EOF;
}

pfi au_header(void)
{
    struct au_header {
		char sndt[4];
		long offset;
		long length;
		long coding;
		long rate;
		long channels;
    } header;
    long ntohl(long);
#ifdef __386__
#   pragma aux ntohl parm [eax] = "xchg ah,al" "rol  eax,16" "xchg ah,al"
#else
#   pragma aux ntohl parm [dx ax] = "xchg dh,al" "xchg dl,ah"
#endif
    pfi fetch;

    xrewind();

    xread(&header, sizeof header);

    if (strncmp(header.sndt, ".snd", sizeof header.sndt) != 0)
		return 0;

    header.offset = ntohl(header.offset);
    header.length = ntohl(header.length);
    header.coding = ntohl(header.coding);
    header.rate   = ntohl(header.rate);

    /* ulaw-8 or linear-8 */
    switch (header.coding) {
      case 1:	/* ulaw-8    */
		fetch = auu_read;
//		samplebits = 16;
		samplebits = 8;
		if (ioctl(fileno(audio), SOUND_PCM_WRITE_BITS, &samplebits) == -1)
		    samplebits = 8;
		break;
      case 2:	/* linear-8  */
 		fetch = aul_read;
		samplebits = 8;
		break;
      case 3:	/* linear-16 */
  		fetch = aul_read;
		samplebits = 16;
		break;
      default:
  		fprintf(stderr, "not ulaw-8 or linear-8\n");
		return 0;
    }

    xskip(header.offset - sizeof header);

    samples = header.length;

    if (samples == 0) samples = -1;

    samplerate = header.rate;
    samplechannels = header.channels;

    if(ioctl(fileno(audio), SOUND_PCM_WRITE_RATE, &samplerate) == -1)
		perror("ioctl(SOUND_PCM_WRITE_RATE)");

	if(samplechannels < 3) {
	    if(ioctl(fileno(audio), SOUND_PCM_WRITE_CHANNELS, &samplechannels) == -1)
			perror("ioctl(SOUND_PCM_WRITE_CHANNELS)");
	}

    if(ioctl(fileno(audio), SOUND_PCM_WRITE_BITS, &samplebits) == -1)
		perror("ioctl(SOUND_PCM_WRITE_BITS)");

    return fetch;
}

voc_read(void)
{
    while (samples == 0) {
		static off_t repeat;
		static int count;
		int type = xgetc();

		if (xread(&samples, 3) == 0)
		    break;

		switch (type) {
		  case 1: {
		    /*-
		     * The voice data sub-block.
		     *
		     * The sampling rate field (sr) is defined as:
		     *    sr = 256 - 1000000 / Hz
		     * so 10,000 Hz is 156.
		     *
		     * The packing types are:
		     *    0 : 8-bit raw or unpacked
		     *    1 : 4-bit packed
		     *    2 : 2.6-bit packed
		     *    3 : 2-bit packed
		     *
		     * The voice data length is calculated as blklen - 2
		     * to account for sr and pack.
		     */
		    samplerate = 1000000L / (256 - (unsigned) xgetc());
		    if (xgetc()) {
				fprintf(stderr, "not pack type 0\n");
				return EOF;
		    }
		    if(ioctl(fileno(audio), SOUND_PCM_WRITE_RATE, &samplerate) == -1)
				perror("ioctl()");
		    samples -= 2;
		    break;
		  }

		  case 2: {
		    /*-
		     * Voice continuation carries on from the last voice data sub-block.
		     */
		    break;
		  }

		  case 3: {
		    /*-
		     * The silence sub-block.
		     *
		     * The period field is one less than the silence period
		     * (i.e. 0 = 1 period)
		     * The blklen is always 3.
		     */
		    struct silence {
				u_short	period;		/* Silence period */
				u_char	sr;		/* Sampling rate */
		    };

		    xgetc(), xgetc(), xgetc();
		    samples -= 3;
		    break;
		  }

  		  case 4: {
		    /*-
		     * The marker sub-block.
		     * Can be used for synchronization purposes (multi-media).
		     * The blklen is always 2.
		     */
		    struct marker {
				u_short	marker;		/* Block marker */
		    };

		    xgetc(), xgetc();
		    samples -= 2;
		    break;
		  }

		  case 5: {
		    /*-
		     * The embedded ascii text sub-block.
		     * The text is nul terminated.
		     * Can be used for anything, typically commentary on the voice data.
		     */
		    xskip(samples);
		    samples = 0;
		    break;
		  }

		  case 6: {
		    /*-
		     * Beginning of a repeat loop.
		     * Play the data sub-blocks from here to the next end repeat loop
		     * (count + 1) times.
		     * The blklen is always 2.
		     */
		    struct repeat_loop {
				u_short	count;		/* Repeat count */
	   		};

		    xread(&count, 2);
		    repeat = ftell(stdin);
		    samples -= 2;
		    break;
		  }

		  case 7: {
		    /*-
		     * End of a repeat loop.
		     * The blklen is always 0.
		     */
		    if (count--)	samples = fseek(stdin, repeat, SEEK_SET);
		    break;
		  }

		  default: {
		    /*-
		     * Unknown -- ATI SFX utility MW2CV uses type 8
		     */
		    break;
		  }
		}
	 }
	    if (samples == 0 || samples--)	return xgetc();

    return EOF;
}

pfi voc_header(void)
{
    struct voc_header {
		char  cvf[20];
		short offset;
		short vers1, vers2;
    } header;

    xrewind();
    xread(&header, sizeof header);

    if (strncmp(header.cvf, "Creative Voice File\032", sizeof header.cvf) != 0)
		return 0;

    xskip(header.offset - sizeof header);

    samples = 0;
    samplerate = 0;
    samplebits = 8;

    if(ioctl(fileno(audio), SOUND_PCM_WRITE_BITS, &samplebits) == -1)
		perror("ioctl()");

    return voc_read;
}

wa4_read(void)
{
    static int cont, ch;

    if ((++cont & 1) == 0)	return (ch & 0xf0) >> 2;

    if ((ch = xgetc()) == EOF) {
		cont = 0;
		return EOF;
    }

    return (ch & 0x0f) << 2;
}

wav_read(void)
{
    if (samples == 0 || samples--)	return xgetc();

    return EOF;
}

pfi wav_header(void)
{
    union {
		struct riff_header {
		    char riff[4];
		    long size;
		    char wave[4];
		} r;
		struct wave_header {
		    short FormatTag;
		    short Channels;
		    long  SamplesPerSec;
		    long  AvgBytesPerSec;
		    short BlockAlign;
		    short BitsPerSample;
		} w;
    } header;

    xrewind();

    if (xread(&header.r, sizeof header.r) == 0)	return 0;

    if (strncmp(header.r.riff, "RIFF", 4) != 0 || strncmp(header.r.wave, "WAVE", 4) != 0)
		return 0;

    samples = xtag("fmt ");

    xread(&header.w, sizeof header.w);
    xskip(samples - sizeof header.w);

    samples = xtag("data");
    samplerate = header.w.SamplesPerSec;
    samplechannels = header.w.Channels;
    samplebits = header.w.BitsPerSample;

    if(ioctl(fileno(audio), SOUND_PCM_WRITE_RATE, &samplerate) == -1)
		perror("ioctl()");

    if(ioctl(fileno(audio), SOUND_PCM_WRITE_CHANNELS, &samplechannels) == -1)
		perror("ioctl()");

    if (samplebits == 4) {
		samples *= 2;
		samplebits = 8;

		if(ioctl(fileno(audio), SOUND_PCM_WRITE_BITS, &samplebits) == -1)
			perror("ioctl()");

		return wa4_read;
    }

    if(ioctl(fileno(audio), SOUND_PCM_WRITE_BITS, &samplebits) == -1)
		perror("ioctl()");

    return wav_read;
}

play(void)
{
    pfi fetch;
    size_t i=0,ch;
	size_t	pos;

    if ((fetch = au_header()) || (fetch = voc_header()) || (fetch = wav_header()))
	;
    else
		return;

    while ((ch = fetch()) != EOF) {
		if (fputc(ch, audio) == -1) {
		    fprintf(stderr, "play: write failed %s\n", strerror(errno));
		    break;
		}
		if((i++%1024) == 0) {
			if((pos=ftell(stdin)) == -1) {
				fprintf(stderr,"tell failed: %s\n", strerror(errno));
				break;
			}
   			fprintf(stderr,"%3d%% (%7ld of %8ld)\r",(pos*100)/playsize,
												   pos, playsize);
		}
	}
}

main(int argc, char *argv[])
{
// setvbuf(stdin, 0, _IOFBF, 8192);
   setvbuf(stdin, 0, _IOFBF, 256);

  if (--argc) {
	while (*++argv) {

	    if ((audio = fopen(AUDIO_DEVICE, "w")) == 0) {
			fprintf(stderr, "play: open %s failed %s\n", AUDIO_DEVICE, strerror(errno));
			exit(1);
   		}

	//    setvbuf(audio, 0, _IOFBF, 8192);
	    setvbuf(audio, 0, _IOFBF, 256);

	    puts(*argv);
	    freopen(*argv, "r", stdin);
		playsize=filesize(stdin);
	    play();
		printf("\n");
	    fclose(audio);
	}
  } else {
	    if ((audio = fopen(AUDIO_DEVICE, "w")) == 0) {
			fprintf(stderr, "play: open %s failed %s\n", AUDIO_DEVICE, strerror(errno));
			exit(1);
   		}
		play();
	    fclose(audio);
  }
}

static size_t filesize(FILE *fp)
{
	size_t	save_pos, size_of_file;

	save_pos=ftell(fp);
	fseek(fp,0L,SEEK_END);
	size_of_file=ftell(fp);
	fseek(fp,save_pos,SEEK_SET);
	return(size_of_file);

}
